package gs.test.v0.api

import cats.effect.Sync
import scala.reflect.ClassTag

/** Opaque type used to check candidate values against expected values.
  */
opaque type Check[A] = A

object Check:

  /** Instantiate a new Check.
    *
    * @param candidate
    *   The value to check.
    * @return
    *   The new [[Check]] instance.
    */
  def apply[A](candidate: A): Check[A] = candidate

  extension [A: ClassTag](candidate: Check[A])
    /** @return
      *   The unwrapped value of this [[Check]].
      */
    def unwrap(): A = candidate

    /** ## Usage
      *
      * {{{
      *   val x = 10
      *   val result: TestResult = check(x).isEqualTo(10)
      * }}}
      *
      * @param expected
      *   The expected value.
      * @return
      *   Successful test result if the candidate value is equal to the expected
      *   value, an error describing the test failure otherwise.
      */
    def isEqualTo(
      expected: A
    )(
      using
      CanEqual[A, A],
      SourcePosition
    ): TestResult =
      Assertion.IsEqualTo.evaluate(candidate, expected)

    /** ## Usage
      *
      * This is an effectful/lazy version of `isEqualTo`.
      *
      * {{{
      *   val x = 10
      *   val effect: F[TestResult] = check(x).isEqualToF(10)
      * }}}
      *
      * @param expected
      *   The expected value.
      * @return
      *   Effect that when evaluated will produce: successful test result if the
      *   candidate value is equal to the expected value, an error describing
      *   the test failure otherwise.
      */
    def isEqualToF[F[_]: Sync](
      expected: A
    )(
      using
      CanEqual[A, A],
      SourcePosition
    ): F[TestResult] =
      Sync[F].delay(isEqualTo(expected))

  extension (check: Check[Boolean])

    /** ## Usage
      *
      * {{{
      *   val x = false
      *   val result: TestResult = check(x).isTrue()
      * }}}
      *
      * @return
      *   Successful test result if the candidate value is `true`, an error
      *   describing the test failure otherwise.
      */
    def isTrue(
    )(
      using
      SourcePosition
    ): TestResult =
      Assertion.IsTrue.evaluate(check)

    /** ## Usage
      *
      * This is an effectful/lazy version of `isTrue`.
      *
      * {{{
      *   val x = false
      *   val effect: F[TestResult] = check(x).isTrueF()
      * }}}
      *
      * @return
      *   Effect that when evaluated will produce: successful test result if the
      *   candidate value is `true`, an error describing the test failure
      *   otherwise.
      */
    def isTrueF[F[_]: Sync](
    )(
      using
      SourcePosition
    ): F[TestResult] =
      Sync[F].delay(isTrue())

end Check
