package gs.test.v0.api

import cats.Show

/** Opaque type representing some _permanent identifier_. These are
  * user-assigned strings that are expected to _not change over time_ for some
  * test. This allows tests to be deterministically tracked. The only constraint
  * for a permanent identifier is that it must not be blank.
  *
  * ## Uniqueness
  *
  * Permanent identifiers are expected to be unique within the scope of a
  * [[TestSuite]]. This means that two groups within the same suite _may not_
  * contain tests that share a permanent identifier.
  */
opaque type PermanentId = String

object PermanentId:

  /** Instantiate a new [[PermanentId]].
    *
    * @param candidate
    *   The candidate string.
    * @return
    *   The new [[PermanentId]] instance.
    * @throws IllegalArgumentException
    *   If the candidate string is blank.
    */
  def apply(candidate: String): PermanentId =
    if candidate.isBlank() then
      throw new IllegalArgumentException(
        "Permanent Identifiers must be non-blank."
      )
    else candidate

  given CanEqual[PermanentId, PermanentId] = CanEqual.derived

  given Show[PermanentId] = pid => pid

end PermanentId
