package gs.test.v0.runtime.engine

import cats.effect.Async
import cats.syntax.all.*
import gs.test.v0.definition.TestDefinition
import gs.test.v0.definition.TestFailure
import gs.test.v0.definition.TestGroupDefinition
import gs.test.v0.definition.TestSuite
import gs.test.v0.runtime.SuiteExecution
import gs.test.v0.runtime.TestExecution
import gs.timing.v0.Timing
import gs.uuid.v0.UUID
import java.time.Clock
import java.time.Instant
import natchez.EntryPoint
import natchez.Span

final class TestEngine[F[_]: Async](
  val configuration: EngineConfiguration,
  timing: Timing[F],
  suiteExecutionIdGenerator: UUID.Generator,
  testExecutionIdGenerator: UUID.Generator,
  clock: Clock,
  val entryPoint: EntryPoint[F]
):

  def runSuite(
    suite: TestSuite,
    tests: fs2.Stream[F, TestGroupDefinition[F]]
  ): F[SuiteExecution] =
    for
      executedAt <- Async[F].delay(Instant.now(clock))
      stats      <- EngineStats.initialize[F]
      _ <- tests
        .mapAsync(configuration.groupConcurrency.toInt())(runGroup)
        .evalTap(updateGroupStats)
        .evalTap(reportGroup)
        .flatMap(groupResult => fs2.Stream.emits(groupResult.testExecutions))
        .evalTap(updateTestStats)
        .evalMap(reportTestExecution)
        .compile
        .drain
      overallDuration <- stats.overallDuration.get
      countSeen       <- stats.countSeen.get
      countSucceeded  <- stats.countSucceeded.get
      countFailed     <- stats.countFailed.get
      countIgnored    <- stats.countIgnored.get
    yield SuiteExecution(
      id = suiteExecutionIdGenerator.next(),
      name = suite.name,
      documentation = suite.documentation,
      duration = overallDuration,
      countSeen = countSeen,
      countSucceeded = countSucceeded,
      countFailed = countFailed,
      countIgnored = countIgnored,
      executedAt = executedAt
    )

  private def updateGroupStats(groupResult: GroupResult): F[Unit] = ???

  private def updateTestStats(testExecution: TestExecution): F[Unit] = ???

  private def reportGroup(groupResult: GroupResult): F[Unit] = ???

  private def reportTestExecution(testExecution: TestExecution): F[Unit] = ???

  private def runSpan[A](
    name: String,
    root: Span[F],
    f: F[A]
  ): F[A] =
    root.span(name).use(_ => f)

  def runGroup(
    group: TestGroupDefinition[F]
  ): F[GroupResult] =
    entryPoint.root("test-group").use { rootSpan =>
      for
        _ <- rootSpan.put("test_group_name" -> group.name.show)
        _ <- runSpan(
          "before-group",
          rootSpan,
          group.beforeGroup.getOrElse(Async[F].unit)
        )
        stream <- executeGroupTests(group, rootSpan)
        _ <- runSpan(
          "after-group",
          rootSpan,
          group.afterGroup.getOrElse(Async[F].unit)
        )
      yield stream
    }

  private def executeGroupTests(
    group: TestGroupDefinition[F],
    rootSpan: Span[F]
  ): F[GroupResult] =
    rootSpan.span("group").use { groupSpan =>
      for
        traceId    <- rootSpan.traceId.map(parseTraceId)
        timer      <- timing.start()
        executions <- streamGroupTests(group, groupSpan).compile.toList
        elapsed    <- timer.checkpoint()
      yield new GroupResult(
        name = group.name,
        documentation = group.documentation,
        duration = elapsed.duration,
        testExecutions = executions
      )
    }

  private def streamGroupTests(
    group: TestGroupDefinition[F],
    groupSpan: Span[F]
  ): fs2.Stream[F, TestExecution] =
    fs2.Stream
      .emits(group.tests)
      .mapAsync(configuration.testConcurrency.toInt()) { test =>
        for
          testExecutionId <- Async[F].delay(
            TestExecution.Id(testExecutionIdGenerator.next())
          )
          timer   <- timing.start()
          _       <- group.beforeEachTest.getOrElse(Async[F].unit)
          result  <- runSingleTest(testExecutionId, test, groupSpan)
          _       <- group.afterEachTest.getOrElse(Async[F].unit)
          elapsed <- timer.checkpoint()
        yield TestExecution(
          id = testExecutionId,
          permanentId = test.permanentId,
          documentation = test.documentation,
          tags = test.tags,
          markers = test.markers,
          result = result,
          traceId = ???,
          sourcePosition = test.sourcePosition,
          duration = elapsed.duration
        )
      }

  private def runSingleTest(
    testExecutionId: TestExecution.Id,
    test: TestDefinition[F],
    groupSpan: Span[F]
  ): F[Either[TestFailure, Any]] =
    groupSpan.span("test").use { span =>
      for
        // TODO: Constants
        _      <- span.put("test_execution_id" -> testExecutionId.show)
        _      <- span.put("test_name" -> test.name.show)
        result <- test.unitOfWork.work(span)
      yield result
    }

  private def parseTraceId(candidate: Option[String]): UUID = ???
